// -*- C++ -*-
#ifndef AGILE_GENERATOR_HH
#define AGILE_GENERATOR_HH 

#include "AGILe/AGILe.hh"
#include "AGILe/Utils.hh"
#include "AGILe/Particle.hh"
#include "AGILe/GeneratorState.hh"
#include "HepMC/GenEvent.h"

namespace AGILe {


  /// @author Andy Buckley
  class Generator {
  public:

    /// Really just syntactic sugar for bool returns from setParam
    enum Result { SUCCESS=1, FAILURE=0 };


    /// Standard constructor.
    Generator() :
      _initialstateset(false), 
      _initialized(false),
      _versionSet(false) 
    { }
    
    /// Destructor.
    virtual ~Generator() { }


    /// Define initial state.
    void setInitialState(int p1 = PROTON, double e1 = 7000.0, 
                         int p2 = PROTON, double e2 = 7000.0) 
    {
      setGenSpecificInitialState(p1, e1, p2, e2);
      /// @todo Make sure it was successful - exception handling
      _initialstateset = true;
      Beam beam1, beam2;
      beam1.name = p1;
      beam2.name = p2;
      beam1.energy = e1;
      beam2.energy = e2;
      beam1.position = 1;
      beam2.position = 2;
      _myState.setBeams(beam1, beam2);
    }
    

    virtual const GeneratorState& getState() {
      string myName = getName();
      _myState.setName(myName);
      _myState.setVersion(getVersion());
      _myState.setCrossSection(getCrossSection());
      PDF pdf;
      for (vector<Beam>::const_iterator beamIt=_myState.getBeams().begin();
          beamIt != _myState.getBeams().end(); ++beamIt) {
        pdf.particleId = beamIt->name;
        pdf.set = getPDFSet(beamIt->name);
        pdf.member = getPDFMember(beamIt->name);
        pdf.scheme = getPDFScheme(beamIt->name);
        _myState.addPDF(pdf);
      }
      return _myState;
    }


    virtual void setState(const GeneratorState& state) {
      throw runtime_error("This ain't ready yet, it's just here as an interface placeholder. Sorry.");
    }


    /// Get generator name.
    virtual const string getName() const {
      /// @todo This doesn't get used polymophically if the calling function is the base class
      /// @todo Move to a member variable cf. Rivet Projs/Analyses
      return "Generator";
    }


    virtual void setVersion(const string& version) {
      _myVersion = version;
      _versionSet = true;
    }
    

    /// Used as a last resort to determine the generator version.
    /// Ideally a specific generator will know its own version and overwrite 
    /// this.  Some generators (e.g. FHerwig) don't know this.
    virtual const string getVersion() {
      if (!_versionSet) {
        /// @todo This getName() doesn't get used polymophically
        getLog() << Log::ERROR 
                 << "You asked the generator " << getName()
                 << " what version it is, but it doesn't know!\n"
                 << "You should set the version with a call to setVersion(string)" << endl;
        throw runtime_error("Version unknown for generator");
      }
      return _myVersion;
    }
    

    /// Get the PDF set used by the generator, e.g. HWLHAPDF (for herwig)
    /// or one of the default sets.
    virtual string getPDFSet(PdgCode pid) {
      throw runtime_error("Have not implemented getPDFSet in " + getName());
    }


    /// Get the PDF id number for a given particle
    virtual int getPDFMember(PdgCode pid) {
      throw runtime_error("Have not implemented getPDFMember in " + getName());
    }


    /// Get the PDF scheme used by the generator (LHAPDF or other)
    virtual string getPDFScheme(PdgCode pid) const {
      throw runtime_error("Have not implemented getPDFScheme in " + getName());
    }

    
    /// Set the random number generator seed.
    virtual void setSeed(const int value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set seed: generator has already been initialized" << endl;
        throw (runtime_error("Tried to set RNG seed after generator initialization"));
      }
      _myState.addSeed(value);
    }


    /// Pass an int-valued parameter to the generator.
    virtual bool setParam(const string& name, const int& value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }


    /// Pass an int-valued parameter to the generator.
    virtual bool setParam(const string& name, const unsigned int& value) {
      if (_initialized) {
        getLog() << Log::ERROR  
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }
    

    /// Pass an int-valued parameter to the generator.
    virtual bool setParam(const string& name, const long& value) {
      if (_initialized) {
        getLog() << Log::ERROR  
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }


    /// Pass an int-valued parameter to the generator.
    virtual bool setParam(const string& name, const unsigned long& value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }


    /// Pass a double/float-valued parameter to the generator.
    virtual bool setParam(const string& name, const double& value) {
      if (_initialized) {
        getLog() << Log::ERROR  
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization.");
      }
      return setParam(name, toString(value));
    }


    /// Pass a double/float-valued parameter to the generator.
    virtual bool setParam(const string& name, const float& value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }

    
    /// Pass a bool-valued parameter to the generator.
    virtual bool setParam(const string& name, const bool& value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      return setParam(name, toString(value));
    }

    
    /// Pass a string-valued parameter to the generator.
    virtual bool setParam(const string& name, const string& value) {
      if (_initialized) {
        getLog() << Log::ERROR 
                 << "Can't set parameter " << name 
                 << ": generator has already been initialized" << endl;
        throw runtime_error("Tried to set param after generator initialization");
      }
      _myState.setParam(name, value);
      return true;
    }


    /// Run one event.
    virtual void makeEvent(HepMC::GenEvent& evt) {
      if (!_initialized) initialize();
      clearEvent(evt);
      _myState.addEvent();
    }

    
    /// Return the generated cross section in pb.
    virtual const double getCrossSection() {
      getLog() << Log::ERROR 
               << " Have not implemented getCrossSection in generator "
               << getName() << endl;
      throw runtime_error("Tried to access cross-section in Generator base class");
      return _crossSection;
    }


    /// Finalise.
    virtual void finalize() = 0;


  protected:

    /// Get named logger
    Log& getLog() const {
      return Log::getLog("AGILe." + getName());
    }


    /// Set initial state implementation details
    virtual void 
    setGenSpecificInitialState(int p1 = PROTON, double e1 = 7000.0, 
                               int p2 = PROTON, double e2 = 7000.0) = 0;


    /// Initialise.
    virtual void initialize() {
      if (!_initialstateset) setInitialState();
    }

    
    /// Clear the supplied event.
    virtual void clearEvent(HepMC::GenEvent& evt) {
      evt.clear();
    }


  protected:    

    double _crossSection;

    /// Flag for detecting whether the setInitialState method has run.
    bool _initialstateset;

    /// Flag for detecting whether the initialization method has run.
    bool _initialized;

    /// Collection of translated names for particles
    map<PdgCode, string> _particleNames;


  private:
    
    GeneratorState _myState;

    bool _versionSet;

    string _myVersion;
  };


  ////////////////////////////////////////////////


  /// Class factory for dynamic generator loading.
  typedef Generator* (*Creator)();


  /// Class destroyer for dynamic generator loading.
  typedef void Destroyer(Generator*);

}

#endif
