// -*- C++ -*-
#include "AGILe/CCHerwig/CCHerwig.hh"
#include "ThePEG/Repository/EventGenerator.h"
#include "ThePEG/Repository/Repository.h"
#include "ThePEG/Utilities/DynamicLoader.h"
#include "ThePEG/EventRecord/Event.h"
#include "ThePEG/Vectors/HepMCConverter.h"


namespace ThePEG {
  template<> 
  struct HepMCTraits<HepMC::GenEvent> 
    : public HepMCTraitsBase<HepMC::GenEvent, 
                             HepMC::GenParticle, 
                             HepMC::GenVertex, 
                             HepMC::Polarization> { };
}
using namespace ThePEG;


namespace AGILe {
  // Standard constructor
  CCHerwig::CCHerwig() {
    // Set up particle name map
    _particleNames[ELECTRON]    = "e-";
    _particleNames[POSITRON]    = "e+";
    _particleNames[PROTON]      = "p+";
    _particleNames[ANTIPROTON]  = "pbar-";
    // NEUTRON ANTINEUTRON PHOTON MUON ANTIMUON TAU ANTITAU
    // NU_E NU_EBAR NU_MU NU_MUBAR NU_TAU NU_TAUBAR PIPLUS PIMINUS PIZERO
    // PHOTOELECTRON PHOTOPOSITRON PHOTOMUON PHOTOANTIMUON PHOTOTAU PHOTOANTITAU

    DynamicLoader::appendPath(HWLIBDIR);
    //DynamicLoader::appendPath(THEPEGLIBDIR);
    Repository::load(string(HWDATADIR) + "/HerwigDefaults.rpo");

    Repository::exec("mkdir /Agile" , cout);
    Repository::exec("cd /Agile" , cout);

    Repository::exec("create ThePEG::Cuts AgileCuts", cout);
    Repository::exec("create ThePEG::LuminosityFunction AgileLuminosity", cout);
    
  }


  // Set up initial state from supplied params
  void CCHerwig::setGenSpecificInitialState(PdgCode p1, double e1, 
                                            PdgCode p2, double e2) {
    getLog() << Log::INFO << "Setting initial state..." << endl;

    // Beam setups
    Repository::exec("cd /Agile" , cout);
    if ((p1 == ELECTRON || p1 == POSITRON) && (p2 == POSITRON || p2 == ELECTRON)) {
      // e+ e-
      getLog() << Log::DEBUG << "Setting LEP-like initial state" << endl;
      Repository::exec("cp /Herwig/Generators/LEPGenerator AgileGenerator", cout);
    } else if ((p1 == PROTON || p1 == ANTIPROTON) && (p2 == PROTON || p2 == ANTIPROTON)) {
      // p p and p pbar
      getLog() << Log::DEBUG << "Setting LHC/Tevatron-like initial state" << endl;
      Repository::exec("cp /Herwig/Generators/LHCGenerator AgileGenerator", cout);
    } else {
      throw std::runtime_error("Unsupported beam type for " + getName());
    }

    // Make string versions of the energies
    stringstream e1_ss, e2_ss;
    e1_ss << e1;
    e2_ss << e2;

    Repository::exec("set AgileGenerator:EventHandler:LuminosityFunction AgileLuminosity", cout);
    Repository::exec("set AgileLuminosity:BeamEMaxA " + e1_ss.str(), cout);
    Repository::exec("set AgileLuminosity:BeamEMaxB " + e2_ss.str(), cout);
    Repository::exec("set AgileGenerator:EventHandler:BeamA /Herwig/Particles/" + _particleNames[p1], cout);
    Repository::exec("set AgileGenerator:EventHandler:BeamB /Herwig/Particles/" + _particleNames[p2], cout);
    
    _initialstateset = true;
  }
  


  // Run generator initialization
  void CCHerwig::initialize() {
    Generator::initialize();
    getLog() << Log::INFO << "CCHerwig initialising..." << endl;
    
    try {
      Repository::exec("set AgileGenerator:EventHandler:Cuts AgileCuts", cout);

      Repository::exec("set AgileGenerator:RandomNumberGenerator:Seed 31122001", cout);
      Repository::exec("set AgileGenerator:DebugLevel 1", cout);
      Repository::exec("set AgileGenerator:PrintEvent 10", cout);
      Repository::exec("set AgileGenerator:MaxErrors 10000", cout);
      Repository::exec("set AgileGenerator:NumberOfEvents 1000000000", cout);
      
      getLog() << Log::DEBUG << "Getting Herwig++ generator from repo..." << endl;
      EGPtr tmp = BaseRepository::GetObject<EGPtr>("/Agile/AgileGenerator");
      getLog() << Log::DEBUG << "Herwig++ making run..." << endl;
      eg = Repository::makeRun(tmp, "Agile");
      getLog() << Log::DEBUG << "Herwig++ initialising..." << endl;
      eg->initialize();
    }
    catch ( std::exception & e ) {
      getLog() << Log::ERROR << e.what() << endl;
      /// @todo Handle better?
      throw;
    }
    _initialized = true;
    getLog() << Log::INFO << "Herwig++ initialised" << endl;
  }


  void CCHerwig::setSeed(const int seed) {
    Generator::setSeed(seed);
    setParam("AgileGenerator:RandomNumberGenerator:Seed", seed);
  }



  // Set parameters
  bool CCHerwig::setParam(const string& name, const string& value) {
    Generator::setParam(name, value);
    // INFILE is a special param to point at a Herwig++ .in file
    if (name == "INFILE") {
      Repository::read(value, cout);
    } else {
      // Otherwise, translate key=val -> "set key val"
      Repository::exec("set " + name + " " + value, cout);
    }
    return SUCCESS;
  }


  
  // Run the generator for one event.
  void CCHerwig::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    getLog() << Log::DEBUG << "Making event" << endl;
    try {
      if (eg) {
        // Generate an event and convert to a HepMC::GenEvent
        EventPtr event = eg->shoot();
        HepMCConverter<HepMC::GenEvent>::convert(*event, evt, false, GeV, millimeter);
      } else {
        getLog() << Log::ERROR << "eg = nil" << endl;
      }      
    }
    catch ( std::exception & e ) {
      getLog() << Log::ERROR << e.what() << endl;
      /// @todo Handle better?
      throw;
    }
    getLog() << Log::DEBUG << "Event made" << endl;
  }

  

  /// Tidy up, print out run stats, etc.
  void CCHerwig::finalize() {
    getLog() << Log::INFO << "Finalising..." << endl;
    assert(eg);
    eg->finalize();
    Repository::cleanup();
  }


}



// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::CCHerwig(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
