#include "AGILe/AGILe.hh"
#include "AGILe/Generator.hh"
#include <dlfcn.h>
#include <unistd.h>

namespace AGILe {


  // Declare getLibPath function (uses binreloc)
  const string getLibPath();


  namespace Loader {


    Log& getLog() {
      return AGILe::Log::getLog("AGILe.Loader");
    }

    
    // Struct for holding library names and versions
    struct GenLibInfo {
      GenLibInfo(string libname, string pkgname="", string pkgversion="") {
        this->libname = libname;
        this->pkgname = pkgname;
        this->pkgversion = pkgversion;
      }
      string libname, pkgname, pkgversion;
    };


    // Static library handles
    vector<void*> handles;
    void* agilehandle = 0;

    // Libraries required for each known generator.
    typedef map<string, vector<GenLibInfo> > GenLibs;

    // Generator version is kept globally for use by the 
    // Generator::setVersion(...) function.
    string version;


    // Initialise the dynamic loader
    void initialize() {
      // Nothing to do with raw dlopen
    }



    // Make the proper system library name
    string sysLibName(string libname) {
      // Put lib suffix on the end if needed
      if (libname.find(SYSDSO) == string::npos) libname += SYSDSO;
      return libname;
    }



    vector<string> getPlatforms(const GenLibInfo& info) {
      vector<string> platforms;
      if (!info.pkgname.empty()) {
        // If we're not trying to load an AGILe interface...

        // Make a list of platform tags to try
        // This non-standard env variable can be specified by the user.
        const char* envlcg = getenv("LCG_PLATFORM");
        if (envlcg) platforms.push_back(envlcg);

        // CMTCONFIG specifies the LCG platform tag
        /// @todo Actually, it uses a different convention, so let's ditch it
        //const char* envcmt = getenv("CMTCONFIG");
        //if (envcmt) platforms.push_back(envcmt);

        // Also try a fallback of the LCG tag, computed at build time.
        platforms.push_back(LCGTAG);
      }
      // Also try with no platform tag
      platforms.push_back("");

      return platforms;
    }



    vector<string> getLibSearchDirs(const GenLibInfo& info) {
      vector<string> dirs;
      if (info.pkgname.empty()) {
        // If we're trying to load an AGILe interface...

        // Use the AGILe library install path
        dirs.push_back(getLibPath());

        // Use the current dir, too.
        dirs.push_back(".");      

        // And then the user's (non-system) library path
        const char* envld = getenv("LD_LIBRARY_PATH");
        if (envld) dirs += split(envld);

      } else {
        // If we're loading a real generator library...

        // Use the AGILe gen path variable first
        const char* envAGP = getenv("AGILE_GEN_PATH");
        if (envAGP) {
          const vector<string> genpaths = split(envAGP);
          foreach (const string& dir, genpaths) {
            dirs.push_back(dir);
            dirs.push_back(dir + "/MCGenerators");
          }
        }

        // Try the Genser area on CERN AFS 
        const char* envAUA = getenv("AGILE_USE_AFS");
        if (!envAUA || toUpper(string(envAUA)) != "NO") {
          dirs.push_back("/afs/cern.ch/sw/lcg/external/MCGenerators");
        }
      }

      return dirs;
    }



    string findGenLib(const GenLibInfo& info, 
                      const vector<string> searchdirs,
                      const vector<string> platforms) 
    {
      // Get platform-specific filename
      string libname = sysLibName(info.libname);
      
      // Try to find lib file here
      foreach (const string& d, searchdirs) {
        foreach (const string& p, platforms) {
          vector<string> versionsuffixes;
          versionsuffixes += ".2","";
          foreach (string& vsuffix, versionsuffixes) {
            string libpath = d + "/" + info.pkgname;
            if (!info.pkgversion.empty()) {
              libpath += "/" + info.pkgversion + vsuffix;
            }
            if (!p.empty()) {
              libpath += "/" + p;
            }
            if (!info.pkgname.empty()) {
              libpath += "/lib";
            }
            libpath += "/" + libname;
            // Remove duplicate slashes from path
            while (libpath.find("//") != string::npos) {
              libpath.replace(libpath.find("//"), 2, "/");
            }
            // Return path if present
            getLog() << Log::TRACE << "Testing for " << libpath << endl;
            if (access(libpath.c_str(), R_OK) == 0) {
              getLog() << Log::TRACE << "Found " << libpath << endl;
              return libpath;
            }
          }
        }
      }
      // Fall back to non-Genser structure
      foreach (const string& d, searchdirs) {
        const string libdir = d + "/lib";
        vector<string> libnames;
        libnames += libname,(info.libname+info.pkgversion+SYSDSO);
        foreach (string& ln, libnames) {
          string libpath = libdir + "/" + ln;
          // Remove duplicate slashes from path
          while (libpath.find("//") != string::npos) {
            libpath.replace(libpath.find("//"), 2, "/");
          }
          // Return path if present
          getLog() << Log::TRACE << "Testing for " << libpath << endl;
          if (access(libpath.c_str(), R_OK) == 0) {
            getLog() << Log::TRACE << "Found " << libpath << endl;
            return libpath;
          }
        }
      }
      return "";
    }



    string findGenLib(const GenLibInfo& info) {
      // Build up lists of search directories and platforms
      vector<string> platforms = getPlatforms(info);
      vector<string> dirs = getLibSearchDirs(info);

      return findGenLib(info, dirs, platforms);
    }



    void* loadGenLib(const string& libpath, int dlmode=(RTLD_LAZY|RTLD_GLOBAL)) {
      getLog() << Log::TRACE << "Trying to load " << libpath << endl;
      void* lib = dlopen(libpath.c_str(), dlmode);
      if (lib) {
        getLog() << Log::TRACE << "Successfully loaded " << libpath << " (" << lib << ")" << endl;
        // If this is an AGILe interface, set the global handle
        if (libpath.find("libAGILe") != string::npos) {
          getLog() << Log::TRACE << "Setting AGILe module handle for " << libpath 
                   << " (" << lib << ")" << endl;
          agilehandle = lib;
        }
      } else {
        getLog() << Log::TRACE << "Failed to load " << libpath << endl;
      }
      return lib;
    }


    void* loadGenLib(const GenLibInfo& info, int dlmode=(RTLD_LAZY|RTLD_GLOBAL)) {
      const string libpath = findGenLib(info);
      if (libpath != "") {
        return loadGenLib(libpath, dlmode);
      }
      return 0;
    }


    GenLibs _getAllGenLibInfo() {
      vector<string> versions;
      GenLibs gens;

      // Is LHAPDF available?
      versions += "5.5.0","5.4.1","5.4.0","5.3.1","5.3.0";
      bool haveLHAPDF = false;
      GenLibInfo lhapdfLibInfo("DUMMY");
      foreach (const string& v, versions) {
        lhapdfLibInfo = GenLibInfo("libLHAPDF", "lhapdf", v);
        haveLHAPDF = ! findGenLib(lhapdfLibInfo).empty();
        if (haveLHAPDF) break;
      }
      // Try to fall back to non-Genser packaging
      if (!haveLHAPDF) {
        lhapdfLibInfo = GenLibInfo("libLHAPDF");
        haveLHAPDF = ! findGenLib(lhapdfLibInfo).empty();
      }
      // Announce the result
      if (haveLHAPDF) {
        Log::getLog("AGILe.Loader") << Log::DEBUG << "LHAPDF is available" << endl;
      } else {
        Log::getLog("AGILe.Loader") << Log::WARN << "LHAPDF is not available" << endl;
      }


      // Fortran Herwig + variants
      versions.clear();
      versions += "6507","6510";
      foreach (const string& v, versions) {
        vector<GenLibInfo> baselibs, reqlibs;
        baselibs.push_back(GenLibInfo("libherwig", "herwig", v));
        if (haveLHAPDF) {
          baselibs.push_back(lhapdfLibInfo);
        } else {
          baselibs.push_back(GenLibInfo("libherwig_pdfdummy", "herwig", v));
        }
        baselibs.push_back(GenLibInfo("libHepMCfio"));
        baselibs.push_back(GenLibInfo("libAGILeFHerwig"));
        const GenLibInfo libinfo_herwigdummy("libherwig_dummy", "herwig", v);
        const GenLibInfo libinfo_jimmy("libjimmy", "jimmy", "4.31");
        const string alpgen_version = "2.1.3d";
        const GenLibInfo libinfo_alpgen("libalpgen", "alpgen", alpgen_version);
        const GenLibInfo libinfo_alpsho("libalpsho", "alpgen", alpgen_version);
        const GenLibInfo libinfo_alpher("libatoher", "alpgen", alpgen_version);
        const GenLibInfo libinfo_charybher("libcharybdis", "charybdis", "1.003h");
        ////
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_herwigdummy);
        gens["Herwig:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_herwigdummy);
        reqlibs.push_back(libinfo_jimmy);
        reqlibs.push_back(GenLibInfo("libAGILeFHerwigJimmy"));
        gens["HerwigJimmy:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_alpgen);
        reqlibs.push_back(libinfo_alpsho);
        reqlibs.push_back(libinfo_alpher);
        reqlibs.push_back(GenLibInfo("libAGILeAlpGenFHerwig"));
        gens["AlpGenHerwig:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_alpgen);
        reqlibs.push_back(libinfo_alpsho);
        reqlibs.push_back(libinfo_alpher);
        reqlibs.push_back(libinfo_jimmy);
        reqlibs.push_back(GenLibInfo("libAGILeFHerwigJimmy"));
        /// @todo Need special libAGILeAlpGenFHerwigJimmy
        reqlibs.push_back(GenLibInfo("libAGILeAlpGenFHerwig"));
        gens["AlpGenHerwigJimmy:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_charybher);
        reqlibs.push_back(GenLibInfo("libAGILeCharybdisFHerwig"));
        gens["CharybdisHerwig:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(libinfo_charybher);
        reqlibs.push_back(libinfo_jimmy);
        reqlibs.push_back(GenLibInfo("libAGILeFHerwigJimmy"));
        reqlibs.push_back(GenLibInfo("libAGILeCharybdisFHerwig"));
        gens["CharybdisHerwigJimmy:" + v] = reqlibs;
      }


      // Fortran Pythia + variants
      versions.clear();
      versions += "206","326","403","404","406","408","411","412","413","414","415","416","417","418";
      foreach (const string& v, versions) {
        vector<GenLibInfo> baselibs, reqlibs;
        baselibs.push_back(GenLibInfo("libpythia6", "pythia6", v));
        if (haveLHAPDF) {
          baselibs.push_back(lhapdfLibInfo);
        } else {
          baselibs.push_back(GenLibInfo("libpythia6_pdfdummy", "pythia6", v));
        }
        baselibs.push_back(GenLibInfo("libHepMCfio"));
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libpythia6_dummy", "pythia6", v));
        reqlibs.push_back(GenLibInfo("libAGILeFPythia"));
        gens["Pythia6:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        const string alpgen_version = "2.1.3d";
        reqlibs.push_back(GenLibInfo("libalpgen", "alpgen", alpgen_version));
        reqlibs.push_back(GenLibInfo("libalpsho", "alpgen", alpgen_version));
        reqlibs.push_back(GenLibInfo("libatopyt", "alpgen", alpgen_version));
        reqlibs.push_back(GenLibInfo("libpythia6_dummy", "pythia6", v)); 
        reqlibs.push_back(GenLibInfo("libAGILeAlpGenFPythia"));
        gens["AlpGenPythia6:" + v] = reqlibs;
        //
        reqlibs = baselibs;
        reqlibs.push_back(GenLibInfo("libcharybdis", "charybdis", "1.003"));
        reqlibs.push_back(GenLibInfo("libpythia6_dummy", "pythia6", v));
        reqlibs.push_back(GenLibInfo("libAGILeCharybdisFPythia"));
        gens["CharybdisPythia6:" + v] = reqlibs;
      }


      #ifdef HAVE_SHERPA
      // Sherpa
      /// @todo Handle multiple versions?
      versions.clear();
      versions += "1.1.2";
      foreach (const string& v, versions) {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("SHERPA-MC/libSherpaMain", "sherpa", v));
        if (haveLHAPDF) {
          reqlibs.push_back(lhapdfLibInfo);
        }
        reqlibs.push_back(GenLibInfo("libAGILeSherpa"));
        gens["Sherpa"] = reqlibs;
        break; //< @todo Handle multiple versions?
      }
      #endif


      #ifdef HAVE_CCPYTHIA
      // Pythia 8
      versions.clear();
      versions += "108","107";
      foreach (const string& v, versions) {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("libpythia8", "pythia8", v));
        reqlibs.push_back(GenLibInfo("libhepmcinterface", "pythia8", v));
        if (haveLHAPDF) {
          reqlibs.push_back(lhapdfLibInfo);
        }
        reqlibs.push_back(GenLibInfo("libAGILeCCPythia"));
        gens["Pythia8:" + v] = reqlibs;
      }

      #endif


      #ifdef HAVE_CCHERWIG
      // Herwig++
      /// @todo Handle multiple versions?
      versions.clear();
      versions += "2.2.1","2.2.0";
      foreach (const string& v, versions) {
        vector<GenLibInfo> reqlibs;
        reqlibs.push_back(GenLibInfo("Herwig++/Herwig", "herwig++", v));
        if (haveLHAPDF) {
          reqlibs.push_back(lhapdfLibInfo);
        }
        reqlibs.push_back(GenLibInfo("libAGILeCCHerwig"));
        gens["Herwig++"] = reqlibs;
        break; //< @todo Handle multiple versions?
      }
      versions.clear();
      #endif

      return gens;
    }


    // Find which generators are present by testing the files.
    vector<string> getAvailableGens() {
      GenLibs genlibs = _getAllGenLibInfo();
      vector<string> availableGens;
      for (GenLibs::const_iterator g = genlibs.begin(); g != genlibs.end(); ++g) {
        getLog() << Log::TRACE << "Trying to find libs for " << g->first << endl;
        // All listed libraries must be available for each generator to be considered useable.
        bool ok = true;
        for (vector<GenLibInfo>::const_iterator l = g->second.begin(); l != g->second.end(); ++l) {
          const string libpath = findGenLib(*l);
          if (libpath != "") {
            getLog() << Log::TRACE << "Found " << l->libname << " at " << libpath << endl;
          } else {
            getLog() << Log::TRACE << "Didn't find " 
                     << l->pkgname << ":" << l->pkgversion << ":" << l->libname << endl;
            ok = false;
            break;
          }
        }
        // If all libs were present, add this gen to the "available" list.
        if (ok) { 
          getLog() << Log::TRACE << "Found all libs for " << g->first << endl;
          availableGens.push_back(g->first); 
        }
      }
      return availableGens;
    }


    void loadGenLibs(const string& genname) {
      const string fioName = "libHepMCfio";
      string agileName = "";
      string pdflibName = "";

      // Get version string from colon separated version string if given.
      string name = genname;
      version = "";
      const size_t colonpos = genname.rfind(":");
      if (colonpos != string::npos) {
        const size_t versionlength = genname.length() - colonpos - 1;
        name = genname.substr(0, colonpos);
        version = genname.substr(colonpos+1, versionlength);
      }

      GenLibs genlibs = _getAllGenLibInfo();
      if (genlibs.find(genname) == genlibs.end()) {
        throw runtime_error("Trying to load an unknown generator combination: " + genname);
      } else {
        const vector<GenLibInfo> libs = genlibs[genname];
        for (vector<GenLibInfo>::const_iterator l = libs.begin(); l != libs.end(); ++l) {
          handles.push_back(loadGenLib(*l));
        }
      }

      // Check and reset dlopen errors
      for (vector<void*>::const_iterator h = handles.begin(); h != handles.end(); ++h) {
        if (*h == 0) {
          throw runtime_error(string("Failed to load libraries: ") + dlerror());
        }
      }
      dlerror();
    }


    template <typename Fn>
    inline Fn evil_cast(void* ptr) {
      return reinterpret_cast<Fn>(reinterpret_cast<size_t>(ptr));
    }


    Generator* createGen() {
      // Load the factory and create an instance of the class
      void* makeGen = dlsym(agilehandle, "create");
      if (char* err = dlerror()) {
        throw runtime_error((string("Can't load creation factory: ") + err).c_str());
      }
      
      // Make the generator object.
      Generator* gen = evil_cast<Creator>(makeGen)();

      // Set the generator version for e.g. HepML dumps of generator state.
      gen->setVersion(version);
      
      return gen;
    }


    void destroyGen(Generator* gen) {
      //Destroyer* destroy_gen = (Destroyer*) dlsym(agilemod, "destroy");
      //if (dlerror()) {
      //  throw runtime_error("Can't load destruction factory");
      //}
      //destroy_gen(gen);
      delete gen;
    }


    // Close the libraries
    void finalize() {
      for (vector<void*>::reverse_iterator h = handles.rbegin(); h != handles.rend(); ++h) {
        if (*h != 0) dlclose(*h);
      }
    }


  }
}
